var __defProp = Object.defineProperty;
var __export = (target, all) => {
  for (var name in all)
    __defProp(target, name, { get: all[name], enumerable: true });
};

// src/debug.ts
var QTS_DEBUG = Boolean(typeof process === "object" && process.env.QTS_DEBUG);
var debugLog = QTS_DEBUG ? console.log.bind(console) : () => {
};

// src/errors.ts
var errors_exports = {};
__export(errors_exports, {
  QuickJSAsyncifyError: () => QuickJSAsyncifyError,
  QuickJSAsyncifySuspended: () => QuickJSAsyncifySuspended,
  QuickJSMemoryLeakDetected: () => QuickJSMemoryLeakDetected,
  QuickJSNotImplemented: () => QuickJSNotImplemented,
  QuickJSUnwrapError: () => QuickJSUnwrapError,
  QuickJSUseAfterFree: () => QuickJSUseAfterFree,
  QuickJSWrongOwner: () => QuickJSWrongOwner
});
var QuickJSUnwrapError = class extends Error {
  constructor(cause, context) {
    super(String(cause));
    this.cause = cause;
    this.context = context;
    this.name = "QuickJSUnwrapError";
  }
};
var QuickJSWrongOwner = class extends Error {
  constructor() {
    super(...arguments);
    this.name = "QuickJSWrongOwner";
  }
};
var QuickJSUseAfterFree = class extends Error {
  constructor() {
    super(...arguments);
    this.name = "QuickJSUseAfterFree";
  }
};
var QuickJSNotImplemented = class extends Error {
  constructor() {
    super(...arguments);
    this.name = "QuickJSNotImplemented";
  }
};
var QuickJSAsyncifyError = class extends Error {
  constructor() {
    super(...arguments);
    this.name = "QuickJSAsyncifyError";
  }
};
var QuickJSAsyncifySuspended = class extends Error {
  constructor() {
    super(...arguments);
    this.name = "QuickJSAsyncifySuspended";
  }
};
var QuickJSMemoryLeakDetected = class extends Error {
  constructor() {
    super(...arguments);
    this.name = "QuickJSMemoryLeakDetected";
  }
};

// src/asyncify-helpers.ts
function* awaitYield(value) {
  return yield value;
}
function awaitYieldOf(generator) {
  return awaitYield(awaitEachYieldedPromise(generator));
}
var AwaitYield = awaitYield;
AwaitYield.of = awaitYieldOf;
function maybeAsyncFn(that, fn) {
  return (...args) => {
    const generator = fn.call(that, AwaitYield, ...args);
    return awaitEachYieldedPromise(generator);
  };
}
function maybeAsync(that, startGenerator) {
  const generator = startGenerator.call(that, AwaitYield);
  return awaitEachYieldedPromise(generator);
}
function awaitEachYieldedPromise(gen) {
  function handleNextStep(step) {
    if (step.done) {
      return step.value;
    }
    if (step.value instanceof Promise) {
      return step.value.then(
        (value) => handleNextStep(gen.next(value)),
        (error) => handleNextStep(gen.throw(error))
      );
    }
    return handleNextStep(gen.next(step.value));
  }
  return handleNextStep(gen.next());
}

// src/lifetime.ts
var Lifetime = class _Lifetime {
  /**
   * When the Lifetime is disposed, it will call `disposer(_value)`. Use the
   * disposer function to implement whatever cleanup needs to happen at the end
   * of `value`'s lifetime.
   *
   * `_owner` is not used or controlled by the lifetime. It's just metadata for
   * the creator.
   */
  constructor(_value, copier, disposer, _owner) {
    this._value = _value;
    this.copier = copier;
    this.disposer = disposer;
    this._owner = _owner;
    this._alive = true;
    this._constructorStack = QTS_DEBUG ? new Error("Lifetime constructed").stack : void 0;
  }
  get alive() {
    return this._alive;
  }
  /**
   * The value this Lifetime protects. You must never retain the value - it
   * may become invalid, leading to memory errors.
   *
   * @throws If the lifetime has been [[dispose]]d already.
   */
  get value() {
    this.assertAlive();
    return this._value;
  }
  get owner() {
    return this._owner;
  }
  get dupable() {
    return !!this.copier;
  }
  /**
   * Create a new handle pointing to the same [[value]].
   */
  dup() {
    this.assertAlive();
    if (!this.copier) {
      throw new Error("Non-dupable lifetime");
    }
    return new _Lifetime(
      this.copier(this._value),
      this.copier,
      this.disposer,
      this._owner
    );
  }
  consume(map) {
    this.assertAlive();
    const result = map(this);
    this.dispose();
    return result;
  }
  /**
   * Dispose of [[value]] and perform cleanup.
   */
  dispose() {
    this.assertAlive();
    if (this.disposer) {
      this.disposer(this._value);
    }
    this._alive = false;
  }
  assertAlive() {
    if (!this.alive) {
      if (this._constructorStack) {
        throw new QuickJSUseAfterFree(
          `Lifetime not alive
${this._constructorStack}
Lifetime used`
        );
      }
      throw new QuickJSUseAfterFree("Lifetime not alive");
    }
  }
};
var StaticLifetime = class extends Lifetime {
  constructor(value, owner) {
    super(value, void 0, void 0, owner);
  }
  // Static lifetime doesn't need a copier to be copiable
  get dupable() {
    return true;
  }
  // Copy returns the same instance.
  dup() {
    return this;
  }
  // Dispose does nothing.
  dispose() {
  }
};
var WeakLifetime = class extends Lifetime {
  constructor(value, copier, disposer, owner) {
    super(value, copier, disposer, owner);
  }
  dispose() {
    this._alive = false;
  }
};
function scopeFinally(scope, blockError) {
  let disposeError;
  try {
    scope.dispose();
  } catch (error) {
    disposeError = error;
  }
  if (blockError && disposeError) {
    Object.assign(blockError, {
      message: `${blockError.message}
 Then, failed to dispose scope: ${disposeError.message}`,
      disposeError
    });
    throw blockError;
  }
  if (blockError || disposeError) {
    throw blockError || disposeError;
  }
}
var Scope = class _Scope {
  constructor() {
    this._disposables = new Lifetime(/* @__PURE__ */ new Set());
  }
  /**
   * Run `block` with a new Scope instance that will be disposed after the block returns.
   * Inside `block`, call `scope.manage` on each lifetime you create to have the lifetime
   * automatically disposed after the block returns.
   *
   * @warning Do not use with async functions. Instead, use [[withScopeAsync]].
   */
  static withScope(block) {
    const scope = new _Scope();
    let blockError;
    try {
      return block(scope);
    } catch (error) {
      blockError = error;
      throw error;
    } finally {
      scopeFinally(scope, blockError);
    }
  }
  static withScopeMaybeAsync(_this, block) {
    return maybeAsync(void 0, function* (awaited) {
      const scope = new _Scope();
      let blockError;
      try {
        return yield* awaited.of(block.call(_this, awaited, scope));
      } catch (error) {
        blockError = error;
        throw error;
      } finally {
        scopeFinally(scope, blockError);
      }
    });
  }
  /**
   * Run `block` with a new Scope instance that will be disposed after the
   * block's returned promise settles. Inside `block`, call `scope.manage` on each
   * lifetime you create to have the lifetime automatically disposed after the
   * block returns.
   */
  static async withScopeAsync(block) {
    const scope = new _Scope();
    let blockError;
    try {
      return await block(scope);
    } catch (error) {
      blockError = error;
      throw error;
    } finally {
      scopeFinally(scope, blockError);
    }
  }
  /**
   * Track `lifetime` so that it is disposed when this scope is disposed.
   */
  manage(lifetime) {
    this._disposables.value.add(lifetime);
    return lifetime;
  }
  get alive() {
    return this._disposables.alive;
  }
  dispose() {
    const lifetimes = Array.from(this._disposables.value.values()).reverse();
    for (const lifetime of lifetimes) {
      if (lifetime.alive) {
        lifetime.dispose();
      }
    }
    this._disposables.dispose();
  }
};

// src/deferred-promise.ts
var QuickJSDeferredPromise = class {
  /**
   * Use [[QuickJSContext.newPromise]] to create a new promise instead of calling
   * this constructor directly.
   * @unstable
   */
  constructor(args) {
    /**
     * Resolve [[handle]] with the given value, if any.
     * Calling this method after calling [[dispose]] is a no-op.
     *
     * Note that after resolving a promise, you may need to call
     * [[QuickJSContext.executePendingJobs]] to propagate the result to the promise's
     * callbacks.
     */
    this.resolve = (value) => {
      if (!this.resolveHandle.alive) {
        return;
      }
      this.context.unwrapResult(
        this.context.callFunction(
          this.resolveHandle,
          this.context.undefined,
          value || this.context.undefined
        )
      ).dispose();
      this.disposeResolvers();
      this.onSettled();
    };
    /**
     * Reject [[handle]] with the given value, if any.
     * Calling this method after calling [[dispose]] is a no-op.
     *
     * Note that after rejecting a promise, you may need to call
     * [[QuickJSContext.executePendingJobs]] to propagate the result to the promise's
     * callbacks.
     */
    this.reject = (value) => {
      if (!this.rejectHandle.alive) {
        return;
      }
      this.context.unwrapResult(
        this.context.callFunction(
          this.rejectHandle,
          this.context.undefined,
          value || this.context.undefined
        )
      ).dispose();
      this.disposeResolvers();
      this.onSettled();
    };
    this.dispose = () => {
      if (this.handle.alive) {
        this.handle.dispose();
      }
      this.disposeResolvers();
    };
    this.context = args.context;
    this.owner = args.context.runtime;
    this.handle = args.promiseHandle;
    this.settled = new Promise((resolve) => {
      this.onSettled = resolve;
    });
    this.resolveHandle = args.resolveHandle;
    this.rejectHandle = args.rejectHandle;
  }
  get alive() {
    return this.handle.alive || this.resolveHandle.alive || this.rejectHandle.alive;
  }
  disposeResolvers() {
    if (this.resolveHandle.alive) {
      this.resolveHandle.dispose();
    }
    if (this.rejectHandle.alive) {
      this.rejectHandle.dispose();
    }
  }
};

export {
  debugLog,
  QuickJSUnwrapError,
  QuickJSWrongOwner,
  QuickJSNotImplemented,
  QuickJSAsyncifyError,
  QuickJSAsyncifySuspended,
  QuickJSMemoryLeakDetected,
  errors_exports,
  maybeAsyncFn,
  Lifetime,
  StaticLifetime,
  WeakLifetime,
  Scope,
  QuickJSDeferredPromise
};
//# sourceMappingURL=chunk-BMKPGQ5W.mjs.map