import {
  DefaultIntrinsics,
  Lifetime,
  QuickJSContext,
  QuickJSMemoryLeakDetected,
  QuickJSNotImplemented,
  QuickJSRuntime,
  QuickJSWASMModule,
  Scope,
  applyBaseRuntimeOptions,
  applyModuleEvalRuntimeOptions,
  debugLog,
  evalOptionsToFlags
} from "./chunk-DVITHKWU.mjs";

// src/index.ts
export * from "@jitl/quickjs-ffi-types";

// src/context-asyncify.ts
var QuickJSAsyncContext = class extends QuickJSContext {
  /**
   * Asyncified version of {@link evalCode}.
   */
  async evalCodeAsync(code, filename = "eval.js", options) {
    const detectModule = options === void 0 ? 1 : 0;
    const flags = evalOptionsToFlags(options);
    let resultPtr = 0;
    try {
      resultPtr = await this.memory.newHeapCharPointer(code).consume(
        (charHandle) => this.ffi.QTS_Eval_MaybeAsync(
          this.ctx.value,
          charHandle.value,
          filename,
          detectModule,
          flags
        )
      );
    } catch (error) {
      debugLog("QTS_Eval_MaybeAsync threw", error);
      throw error;
    }
    const errorPtr = this.ffi.QTS_ResolveException(this.ctx.value, resultPtr);
    if (errorPtr) {
      this.ffi.QTS_FreeValuePointer(this.ctx.value, resultPtr);
      return { error: this.memory.heapValueHandle(errorPtr) };
    }
    return { value: this.memory.heapValueHandle(resultPtr) };
  }
  /**
   * Similar to {@link newFunction}.
   * Convert an async host Javascript function into a synchronous QuickJS function value.
   *
   * Whenever QuickJS calls this function, the VM's stack will be unwound while
   * waiting the async function to complete, and then restored when the returned
   * promise resolves.
   *
   * Asyncified functions must never call other asyncified functions or
   * `import`, even indirectly, because the stack cannot be unwound twice.
   *
   * See [Emscripten's docs on Asyncify](https://emscripten.org/docs/porting/asyncify.html).
   */
  newAsyncifiedFunction(name, fn) {
    return this.newFunction(name, fn);
  }
};

// src/esmHelpers.ts
function actualUnwrapDefault(mod) {
  const maybeUnwrap = mod.default;
  return maybeUnwrap ?? mod;
}
var unwrapTypescript = actualUnwrapDefault;

// src/from-variant.ts
async function newQuickJSWASMModuleFromVariant(variantOrPromise) {
  const variant = smartUnwrap(await variantOrPromise);
  const [wasmModuleLoader, QuickJSFFI, { QuickJSWASMModule: QuickJSWASMModule2 }] = await Promise.all([
    variant.importModuleLoader().then(smartUnwrap),
    variant.importFFI(),
    import("./module-I7BJGUHV.mjs").then(unwrapTypescript)
  ]);
  const wasmModule = await wasmModuleLoader();
  wasmModule.type = "sync";
  const ffi = new QuickJSFFI(wasmModule);
  return new QuickJSWASMModule2(wasmModule, ffi);
}
async function newQuickJSAsyncWASMModuleFromVariant(variantOrPromise) {
  const variant = smartUnwrap(await variantOrPromise);
  const [wasmModuleLoader, QuickJSAsyncFFI, { QuickJSAsyncWASMModule: QuickJSAsyncWASMModule2 }] = await Promise.all([
    variant.importModuleLoader().then(smartUnwrap),
    variant.importFFI(),
    import("./module-asyncify-F7TJAE3H.mjs").then(unwrapTypescript)
  ]);
  const wasmModule = await wasmModuleLoader();
  wasmModule.type = "async";
  const ffi = new QuickJSAsyncFFI(wasmModule);
  return new QuickJSAsyncWASMModule2(wasmModule, ffi);
}
function memoizePromiseFactory(fn) {
  let promise;
  return () => {
    return promise ?? (promise = fn());
  };
}
function smartUnwrap(val) {
  if (`default` in val) {
    if (`default` in val.default) {
      return val.default.default;
    }
    return val.default;
  }
  return val;
}

// src/vm-interface.ts
function isSuccess(successOrFail) {
  return "error" in successOrFail === false;
}
function isFail(successOrFail) {
  return "error" in successOrFail === true;
}

// src/interrupt-helpers.ts
function shouldInterruptAfterDeadline(deadline) {
  const deadlineAsNumber = typeof deadline === "number" ? deadline : deadline.getTime();
  return function() {
    return Date.now() > deadlineAsNumber;
  };
}

// src/module-test.ts
var TestQuickJSWASMModule = class {
  constructor(parent) {
    this.parent = parent;
    this.contexts = /* @__PURE__ */ new Set();
    this.runtimes = /* @__PURE__ */ new Set();
  }
  newRuntime(options) {
    const runtime = this.parent.newRuntime({
      ...options,
      ownedLifetimes: [
        new Lifetime(void 0, void 0, () => this.runtimes.delete(runtime)),
        ...options?.ownedLifetimes ?? []
      ]
    });
    this.runtimes.add(runtime);
    return runtime;
  }
  newContext(options) {
    const context = this.parent.newContext({
      ...options,
      ownedLifetimes: [
        new Lifetime(void 0, void 0, () => this.contexts.delete(context)),
        ...options?.ownedLifetimes ?? []
      ]
    });
    this.contexts.add(context);
    return context;
  }
  evalCode(code, options) {
    return this.parent.evalCode(code, options);
  }
  disposeAll() {
    const allDisposables = [...this.contexts, ...this.runtimes];
    this.runtimes.clear();
    this.contexts.clear();
    allDisposables.forEach((d) => {
      if (d.alive) {
        d.dispose();
      }
    });
  }
  assertNoMemoryAllocated() {
    const leaksDetected = this.getFFI().QTS_RecoverableLeakCheck();
    if (leaksDetected) {
      throw new QuickJSMemoryLeakDetected("Leak sanitizer detected un-freed memory");
    }
    if (this.contexts.size > 0) {
      throw new QuickJSMemoryLeakDetected(`${this.contexts.size} contexts leaked`);
    }
    if (this.runtimes.size > 0) {
      throw new QuickJSMemoryLeakDetected(`${this.runtimes.size} runtimes leaked`);
    }
  }
  /** @private */
  getFFI() {
    return this.parent.getFFI();
  }
};

// src/runtime-asyncify.ts
var QuickJSAsyncRuntime = class extends QuickJSRuntime {
  /** @private */
  constructor(args) {
    super(args);
  }
  newContext(options = {}) {
    if (options.intrinsics && options.intrinsics !== DefaultIntrinsics) {
      throw new Error("TODO: Custom intrinsics are not supported yet");
    }
    const ctx = new Lifetime(this.ffi.QTS_NewContext(this.rt.value), void 0, (ctx_ptr) => {
      this.contextMap.delete(ctx_ptr);
      this.callbacks.deleteContext(ctx_ptr);
      this.ffi.QTS_FreeContext(ctx_ptr);
    });
    const context = new QuickJSAsyncContext({
      module: this.module,
      ctx,
      ffi: this.ffi,
      rt: this.rt,
      ownedLifetimes: [],
      runtime: this,
      callbacks: this.callbacks
    });
    this.contextMap.set(ctx.value, context);
    return context;
  }
  setModuleLoader(moduleLoader, moduleNormalizer) {
    super.setModuleLoader(
      moduleLoader,
      moduleNormalizer
    );
  }
  /**
   * Set the max stack size for this runtime in bytes.
   * To remove the limit, set to `0`.
   *
   * Setting this limit also adjusts the global `ASYNCIFY_STACK_SIZE` for the entire {@link QuickJSAsyncWASMModule}.
   * See the [pull request](https://github.com/justjake/quickjs-emscripten/pull/114) for more details.
   */
  setMaxStackSize(stackSize) {
    return super.setMaxStackSize(stackSize);
  }
};

// src/module-asyncify.ts
var QuickJSAsyncWASMModule = class extends QuickJSWASMModule {
  /** @private */
  constructor(module, ffi) {
    super(module, ffi);
    this.ffi = ffi;
    this.module = module;
  }
  /**
   * Create a new async runtime inside this WebAssembly module. All runtimes inside a
   * module are limited to a single async call at a time. For multiple
   * concurrent async actions, create multiple WebAssembly modules.
   */
  newRuntime(options = {}) {
    const rt = new Lifetime(this.ffi.QTS_NewRuntime(), void 0, (rt_ptr) => {
      this.callbacks.deleteRuntime(rt_ptr);
      this.ffi.QTS_FreeRuntime(rt_ptr);
    });
    const runtime = new QuickJSAsyncRuntime({
      module: this.module,
      ffi: this.ffi,
      rt,
      callbacks: this.callbacks
    });
    applyBaseRuntimeOptions(runtime, options);
    if (options.moduleLoader) {
      runtime.setModuleLoader(options.moduleLoader);
    }
    return runtime;
  }
  /**
   * A simplified API to create a new {@link QuickJSAsyncRuntime} and a
   * {@link QuickJSAsyncContext} inside that runtime at the same time. The runtime will
   * be disposed when the context is disposed.
   */
  newContext(options = {}) {
    const runtime = this.newRuntime();
    const lifetimes = options.ownedLifetimes ? options.ownedLifetimes.concat([runtime]) : [runtime];
    const context = runtime.newContext({ ...options, ownedLifetimes: lifetimes });
    runtime.context = context;
    return context;
  }
  /** Synchronous evalCode is not supported. */
  evalCode() {
    throw new QuickJSNotImplemented("QuickJSWASMModuleAsyncify.evalCode: use evalCodeAsync instead");
  }
  /**
   * One-off evaluate code without needing to create a {@link QuickJSAsyncRuntime} or
   * {@link QuickJSAsyncContext} explicitly.
   *
   * This version allows for asynchronous Ecmascript module loading.
   *
   * Note that only a single async action can occur at a time inside the entire WebAssembly module.
   * **Multiple concurrent async actions is an error.**
   *
   * See the documentation for [[QuickJSWASMModule.evalCode]] for more details.
   */
  evalCodeAsync(code, options) {
    return Scope.withScopeAsync(async (scope) => {
      const vm = scope.manage(this.newContext());
      applyModuleEvalRuntimeOptions(vm.runtime, options);
      const result = await vm.evalCodeAsync(code, "eval.js");
      if (options.memoryLimitBytes !== void 0) {
        vm.runtime.setMemoryLimit(-1);
      }
      if (result.error) {
        const error = vm.dump(scope.manage(result.error));
        throw error;
      }
      const value = vm.dump(scope.manage(result.value));
      return value;
    });
  }
};

export {
  QuickJSAsyncContext,
  QuickJSAsyncRuntime,
  QuickJSAsyncWASMModule,
  newQuickJSWASMModuleFromVariant,
  newQuickJSAsyncWASMModuleFromVariant,
  memoizePromiseFactory,
  isSuccess,
  isFail,
  shouldInterruptAfterDeadline,
  TestQuickJSWASMModule
};
//# sourceMappingURL=chunk-XVIOV4LL.mjs.map