import "./chunk-Q2VUCIMT.mjs";
import {
  DefaultIntrinsics,
  QuickJSContext,
  QuickJSRuntime,
  QuickJSWASMModule,
  applyBaseRuntimeOptions,
  applyModuleEvalRuntimeOptions,
  evalOptionsToFlags
} from "./chunk-YR34QNKA.mjs";
import {
  Lifetime,
  QuickJSNotImplemented,
  Scope,
  debugLog
} from "./chunk-ENDFUCZG.mjs";

// src/context-asyncify.ts
var QuickJSAsyncContext = class extends QuickJSContext {
  /**
   * Asyncified version of {@link evalCode}.
   */
  async evalCodeAsync(code, filename = "eval.js", options) {
    const detectModule = options === void 0 ? 1 : 0;
    const flags = evalOptionsToFlags(options);
    let resultPtr = 0;
    try {
      resultPtr = await this.memory.newHeapCharPointer(code).consume(
        (charHandle) => this.ffi.QTS_Eval_MaybeAsync(
          this.ctx.value,
          charHandle.value,
          filename,
          detectModule,
          flags
        )
      );
    } catch (error) {
      debugLog("QTS_Eval_MaybeAsync threw", error);
      throw error;
    }
    const errorPtr = this.ffi.QTS_ResolveException(this.ctx.value, resultPtr);
    if (errorPtr) {
      this.ffi.QTS_FreeValuePointer(this.ctx.value, resultPtr);
      return { error: this.memory.heapValueHandle(errorPtr) };
    }
    return { value: this.memory.heapValueHandle(resultPtr) };
  }
  /**
   * Similar to {@link newFunction}.
   * Convert an async host Javascript function into a synchronous QuickJS function value.
   *
   * Whenever QuickJS calls this function, the VM's stack will be unwound while
   * waiting the async function to complete, and then restored when the returned
   * promise resolves.
   *
   * Asyncified functions must never call other asyncified functions or
   * `import`, even indirectly, because the stack cannot be unwound twice.
   *
   * See [Emscripten's docs on Asyncify](https://emscripten.org/docs/porting/asyncify.html).
   */
  newAsyncifiedFunction(name, fn) {
    return this.newFunction(name, fn);
  }
};

// src/runtime-asyncify.ts
var QuickJSAsyncRuntime = class extends QuickJSRuntime {
  /** @private */
  constructor(args) {
    super(args);
  }
  newContext(options = {}) {
    if (options.intrinsics && options.intrinsics !== DefaultIntrinsics) {
      throw new Error("TODO: Custom intrinsics are not supported yet");
    }
    const ctx = new Lifetime(this.ffi.QTS_NewContext(this.rt.value), void 0, (ctx_ptr) => {
      this.contextMap.delete(ctx_ptr);
      this.callbacks.deleteContext(ctx_ptr);
      this.ffi.QTS_FreeContext(ctx_ptr);
    });
    const context = new QuickJSAsyncContext({
      module: this.module,
      ctx,
      ffi: this.ffi,
      rt: this.rt,
      ownedLifetimes: [],
      runtime: this,
      callbacks: this.callbacks
    });
    this.contextMap.set(ctx.value, context);
    return context;
  }
  setModuleLoader(moduleLoader, moduleNormalizer) {
    super.setModuleLoader(
      moduleLoader,
      moduleNormalizer
    );
  }
  /**
   * Set the max stack size for this runtime in bytes.
   * To remove the limit, set to `0`.
   *
   * Setting this limit also adjusts the global `ASYNCIFY_STACK_SIZE` for the entire {@link QuickJSAsyncWASMModule}.
   * See the [pull request](https://github.com/justjake/quickjs-emscripten/pull/114) for more details.
   */
  setMaxStackSize(stackSize) {
    return super.setMaxStackSize(stackSize);
  }
};

// src/module-asyncify.ts
var QuickJSAsyncWASMModule = class extends QuickJSWASMModule {
  /** @private */
  constructor(module, ffi) {
    super(module, ffi);
    this.ffi = ffi;
    this.module = module;
  }
  /**
   * Create a new async runtime inside this WebAssembly module. All runtimes inside a
   * module are limited to a single async call at a time. For multiple
   * concurrent async actions, create multiple WebAssembly modules.
   */
  newRuntime(options = {}) {
    const rt = new Lifetime(this.ffi.QTS_NewRuntime(), void 0, (rt_ptr) => {
      this.callbacks.deleteRuntime(rt_ptr);
      this.ffi.QTS_FreeRuntime(rt_ptr);
    });
    const runtime = new QuickJSAsyncRuntime({
      module: this.module,
      ffi: this.ffi,
      rt,
      callbacks: this.callbacks
    });
    applyBaseRuntimeOptions(runtime, options);
    if (options.moduleLoader) {
      runtime.setModuleLoader(options.moduleLoader);
    }
    return runtime;
  }
  /**
   * A simplified API to create a new {@link QuickJSAsyncRuntime} and a
   * {@link QuickJSAsyncContext} inside that runtime at the same time. The runtime will
   * be disposed when the context is disposed.
   */
  newContext(options = {}) {
    const runtime = this.newRuntime();
    const lifetimes = options.ownedLifetimes ? options.ownedLifetimes.concat([runtime]) : [runtime];
    const context = runtime.newContext({ ...options, ownedLifetimes: lifetimes });
    runtime.context = context;
    return context;
  }
  /** Synchronous evalCode is not supported. */
  evalCode() {
    throw new QuickJSNotImplemented("QuickJSWASMModuleAsyncify.evalCode: use evalCodeAsync instead");
  }
  /**
   * One-off evaluate code without needing to create a {@link QuickJSAsyncRuntime} or
   * {@link QuickJSAsyncContext} explicitly.
   *
   * This version allows for asynchronous Ecmascript module loading.
   *
   * Note that only a single async action can occur at a time inside the entire WebAssembly module.
   * **Multiple concurrent async actions is an error.**
   *
   * See the documentation for [[QuickJSWASMModule.evalCode]] for more details.
   */
  evalCodeAsync(code, options) {
    return Scope.withScopeAsync(async (scope) => {
      const vm = scope.manage(this.newContext());
      applyModuleEvalRuntimeOptions(vm.runtime, options);
      const result = await vm.evalCodeAsync(code, "eval.js");
      if (options.memoryLimitBytes !== void 0) {
        vm.runtime.setMemoryLimit(-1);
      }
      if (result.error) {
        const error = vm.dump(scope.manage(result.error));
        throw error;
      }
      const value = vm.dump(scope.manage(result.value));
      return value;
    });
  }
};
export {
  QuickJSAsyncWASMModule
};
//# sourceMappingURL=module-asyncify-4V7SENBO.mjs.map